modLoader.define 'adapters/reddit', (require,exports) ->
  # TODO remove dummy data
  #DummyData = modLoader.require 'adapters/dummy_data'

  Deferred  = modLoader.require('libs/deferred')

  SubReddit = modLoader.require('models/sub_reddit')
  Post      = modLoader.require('models/post')
  Comment   = modLoader.require('models/comment')  

  exports.ERROR_TYPES = ERROR_TYPES =
    NO_AUTH: 'User is not authenticated'

  exports.REDDITS_SORT_TYPES = REDDITS_SORT_TYPES =
    POPULAR: 'popular'
    NEW:     'new'
    BANNED:  'banned'

  exports.POST_SORT_TYPES = POST_SORT_TYPES =
    HOT: 'hot'
    NEW: 'new'
    TOP: 'top'
    CONTROVERSIAL: 'controversial'

  exports.OWN_REDDITS_TYPES = OWN_REDDITS_TYPES =
    SUB: 'subscriber'
    CONTRIB: 'contributor'
    MOD:  'moderator'

  exports.PAGINATE_TYPE = PAGINATE_TYPE =
    BEFORE: 'before'
    AFTER:  'after'

  exports.create = (ajax, logger)->
    _reqCounter   = 0;
    _queuedReqIds = [];
    _requests     = {}
    _currentReqId = null;

    _userSession = null

    # Private Methods
    _enqueueRequest = (reqId) ->
      _queuedReqIds.push(reqId)

      if !_currentReqId
        _doNextRequest()
      return

    _buildRequest = (options, df) ->
      request = {}
      reqId = ++_reqCounter;

      root = if options.isSecure then 'https://ssl.reddit.com' else 'http://www.reddit.com'
      request.url = root+options.path
      request.dataType = 'json'
      request.data = {}

      # Adds the options data to the request
      if options.data
        for own key,value of options.data
          request.data[key] = value

      # We need to mark this request as json for both post and get methods
      if options.isPost
        request.type = 'POST'
        request.data.api_type = 'json'
      else
        request.type = 'GET'
        request.url += '/.json'

      # Authenticated request need the user session
      if options.isAuth
        request.data.uh = _userSession

      # Basic error and success handling
      request.success = (response, textStatus, xhr)->
        options.success(response)
        _doNextRequest()
        return

      request.error = (xhr, textStatus, thrown)->
        options.failure({message:'network failed, '+textStatus})
        _doNextRequest()
        return

      _requests[reqId] = request;
      return reqId

    _cancelRequest = (reqId) ->
      return

    _cancelAllRequests = ->
      return

    _doNextRequest = ->
      _currentReqId = _queuedReqIds.shift()
      return if !_currentReqId

      request = _requests[_currentReqId]
      if !request
        logger.error('Tried process request '+_currentReqId+', but it no longer points to a valid object')
        _doNextRequest()
        return

      ajax(request)

      return;
      
    # Converts a response containing a list of reddits
    # to a manageable object
    # @param {Object} Data returned from the request
    _formatRedditListingToArray = (origData)->
      object =
        type: origData.kind
        before: origData.data.before
        after: origData.data.after
        data: []

      for item in origData.data.children
        model = null

        if item.kind == 't5'
          model = SubReddit.create(item.data)
        else if item.kind == 't3'
          model = Post.create(item.data)
        else if item.kind == 't1'
          model = Comment.create(item.data)
        
        object.data.push(model)
      return object

    # Public methodsGET
    isLoggedIn = ->
      return !!_userSession

    userSession = (session)->
      _userSession = session if session != undefined
      return _userSession


    # Stops the current request
    abortCurrent = ->
      _cancelRequest(_currentReqId) if _currentReqId
      return

    # Suspends all requests
    abortAll = ->
      _cancelAllRequests()
      return

    #############################
    # Unauthenticated Commands
    #############################

    # Attempts to log the user in,
    # @param user {String} user name
    # @param pass {String} user password
    # @return Promise
    loginUser = (user,pass)->
      df = Deferred.create()
      options =
        isSecure: true
        isPost: true
        path: '/api/login/'+user
        data:
          user: user
          passwd: pass
        success: (response)->
          _userSession = response.uh
          df.resolve()
          return
        failure: (reason)->
          df.reject(reason)
          return

      reqId = _buildRequest(options, df)
      _enqueueRequest(reqId)
      return df.promise()

    # Gets a list of reddits starting from the next
    # No Auth required
    # @param mod {PAGINATE_TYPE} Value representing which direction we are moving in
    # @param target {String} identifier to be combined with the pagination type for getting a
    #   new page
    # @param type {REDDITS_SORT_TYPES} type of the items to select
    # @return Promise
    listAllSubReddits = (mod, target, type)->
      df = Deferred.create()
      options =
        path: '/reddits/'+type
        data:
          count: 25
        success: (response)->
          reddits = _formatRedditListingToArray(response)
          df.resolve(reddits)
          return
        failure: (reason)->
          df.reject(reason)
          return
      # Add the before/after targets
      options.data[mod] = target

      reqId = _buildRequest(options,df)
      # TODO remove this line when done with dummy data
      #options.success(DummyData.subReddits)
      _enqueueRequest(reqId)
      return df.promise()

    # Attempts to get a list of posts for the sub reddit provided@_postScroller
    # No auth required
    # @param mod {PAGINATE_TYPE} Value representing which direction we are moving in
    # @param target {String} identifier to be combined with the pagination type for getting a
    #   new page
    # @param subReddit {SubReddit} object defining this sub reddit
    # @param type {POST_SORT_TYPES} specifies which type of query to make
    # @return promise
    listSubRedditPost = (mod, target, subReddit, type) =>
      df = Deferred.create()
      options =
        isAuth: if isLoggedIn() then true else false
        isPost: if isLoggedIn() then true else false
        path: subReddit.url()+type
        data:
          count: 25
        success: (response) ->
          posts = _formatRedditListingToArray(response)
          df.resolve(posts)
          return
        failure: (reason) ->
          df.reject(reason)
          return
      # Add the before/after targets
      options.data[mod] = target

      reqId = _buildRequest(options,df)
      # TODO remove this line when done with dummy data
      #options.success(DummyData.posts)
      _enqueueRequest(reqId)
      return df.promise()


    #############################
    # Requires authentication
    #############################

    # Gets a list of the user's subscribed sub reddits
    # Requires Authentication
    # @param mod {PAGINATE_TYPE} Value representing which direction we are moving in
    # @param target {String} identifier to be combined with the pagination type for getting a
    #   new page
    # @param type {OWN_REDDITS_TYPES}
    # @return Promise
    listUserSubReddits = (mod, target, type)->
      df = Deferred.create()
      # Fail the operation off the bat if the user is not logged in.
      if (!isLoggedIn())
        df.reject({code:ERROR_TYPES.NO_AUTH})
        return df.promise()

      options =
        isPost: true
        isAuth: true
        path: '/reddits/mine/'+type
        data:
          count: 25
        success: (response)->
          reddits = _formatRedditListingToArray(response)
          df.resolve(reddits)
          return
        failure: (reason)->
          df.reject(reason)
          return
      # Add the before/after targets
      options.data[mod] = target

      reqId = _buildRequest(options,df)
      _enqueueRequest(reqId)
      return df.promise()
    

    return {
      # Properties
      userSession:     userSession
      # Unauthenticated
      loginUser:         loginUser
      listAllSubReddits: listAllSubReddits
      listSubRedditPost: listSubRedditPost
      #Authenticated
      listUserSubReddits: listUserSubReddits
      # Utils
      isLoggedIn:      isLoggedIn
      abortCurrent:    abortCurrent
      abortAll:        abortAll
      }

  return